"use strict";
/* tslint:disable:max-classes-per-file */
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    }
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var NonEmptyArray_1 = require("fp-ts/lib/NonEmptyArray");
var OptionFn = require("fp-ts/lib/Option");
var ValidationFn = require("fp-ts/lib/Validation");
var Decoder = /** @class */ (function () {
    function Decoder() {
    }
    Decoder.andThen = function (decoderFn, decoder) {
        return new AndThenD(decoderFn, decoder);
    };
    Decoder.andMap = function (decoder) {
        return function (fnInDecoder) {
            return Decoder.map2(pipe, decoder, fnInDecoder);
        };
    };
    Decoder.array = function (decoder) {
        return new ArrayD(decoder);
    };
    Decoder.at = function (keys, decoder) {
        return keys.reduceRight(function (decoders, key) { return Decoder.field(key, decoders); }, decoder);
    };
    Decoder.boolean = function () {
        return new BooleanD();
    };
    Decoder.fail = function (message) {
        return new FailD(message);
    };
    Decoder.field = function (key, decoder) {
        return new FieldD(key, decoder);
    };
    Decoder.index = function (key, decoder) {
        return new IndexD(key, decoder);
    };
    Decoder.map = function (fn, decoder) {
        return new MapD(fn, decoder);
    };
    Decoder.map2 = function (fn, decoderA, decoderB) {
        return new Map2D(fn, decoderA, decoderB);
    };
    Decoder.map3 = function (fn, decoderA, decoderB, decoderC) {
        return new Map3D(fn, decoderA, decoderB, decoderC);
    };
    Decoder.nonEmptyArray = function (decoder) {
        return Decoder.andThen(function (_a) {
            var head = _a[0], tail = _a.slice(1);
            if (head) {
                return Decoder.succeed(new NonEmptyArray_1.NonEmptyArray(head, tail));
            }
            return Decoder.fail('Array must have at least one value');
        }, Decoder.array(decoder));
    };
    Decoder.null = function (value) {
        return new NullD(value);
    };
    Decoder.nullable = function (decoder) {
        var decoders = new NonEmptyArray_1.NonEmptyArray(Decoder.null(OptionFn.none), [Decoder.map(OptionFn.some, decoder)]);
        return Decoder.oneOf(decoders);
    };
    Decoder.number = function () {
        return new NumberD();
    };
    Decoder.oneOf = function (decoders) {
        return new OneOfD(decoders);
    };
    Decoder.optionalField = function (key, decoder) {
        return new OptionalFieldD(key, decoder);
    };
    Decoder.string = function () {
        return new StringD();
    };
    Decoder.succeed = function (value) {
        return new SucceedD(value);
    };
    Decoder.unknown = function () {
        return new UnknownD();
    };
    Decoder.prototype.run = function (value) {
        throw new Error('This method has to be implemented');
    };
    return Decoder;
}());
exports.default = Decoder;
var AndThenD = /** @class */ (function (_super) {
    __extends(AndThenD, _super);
    function AndThenD(decoderFn, decoder) {
        var _this = _super.call(this) || this;
        _this.decoder = decoder;
        _this.decoderFn = decoderFn;
        return _this;
    }
    AndThenD.prototype.run = function (value) {
        var _this = this;
        return this.decoder
            .run(value)
            .fold(function (errors) { return ValidationFn.failure(errors); }, function (success) { return _this.decoderFn(success).run(value); });
    };
    return AndThenD;
}(Decoder));
var ArrayD = /** @class */ (function (_super) {
    __extends(ArrayD, _super);
    function ArrayD(decoder) {
        var _this = _super.call(this) || this;
        _this.decoder = decoder;
        return _this;
    }
    ArrayD.prototype.run = function (valueArray) {
        var _this = this;
        if (isArray(valueArray)) {
            return valueArray.reduce(function (validated, nextValue, indexN) {
                var validatedValue = _this.decoder.run(nextValue);
                return validatedValue.fold(function (errorsFromValue) {
                    var errors = errorsFromValue.map(function (errorText) { return errorText + " at index #" + indexN; });
                    return validated.fold(function (errorsFromValidated) { return ValidationFn.failure(errorsFromValidated.concat(errors)); }, function () { return ValidationFn.failure(errors); });
                }, function (successFromValue) { return validated.map(function (successes) { return successes.concat([successFromValue]); }); });
            }, ValidationFn.success([]));
        }
        var message = "Value must be an array, found \"" + typeof valueArray + "\" instead";
        return ValidationFn.failure(NonEmptyArray_1.nonEmptyArray.of(message));
    };
    return ArrayD;
}(Decoder));
var BooleanD = /** @class */ (function (_super) {
    __extends(BooleanD, _super);
    function BooleanD() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    BooleanD.prototype.run = function (value) {
        if (typeof value === 'boolean') {
            return ValidationFn.success(value);
        }
        var message = "Value must be a boolean, found \"" + typeof value + "\" instead";
        return ValidationFn.failure(NonEmptyArray_1.nonEmptyArray.of(message));
    };
    return BooleanD;
}(Decoder));
var FailD = /** @class */ (function (_super) {
    __extends(FailD, _super);
    function FailD(message) {
        var _this = _super.call(this) || this;
        _this.message = message;
        return _this;
    }
    FailD.prototype.run = function () {
        return ValidationFn.failure(NonEmptyArray_1.nonEmptyArray.of(this.message));
    };
    return FailD;
}(Decoder));
var FieldD = /** @class */ (function (_super) {
    __extends(FieldD, _super);
    function FieldD(field, decoder) {
        var _this = _super.call(this) || this;
        _this.decoder = decoder;
        _this.field = field;
        return _this;
    }
    FieldD.prototype.run = function (valueObject) {
        var _this = this;
        if (isObject(valueObject)) {
            if (this.field in valueObject) {
                return this.decoder
                    .run(valueObject[this.field])
                    .mapFailure(function (errors) { return errors.map(function (error) { return error + " at key \"" + _this.field + "\""; }); });
            }
            var noFieldMsg = "Object missing value at key \"" + this.field + "\"";
            return ValidationFn.failure(NonEmptyArray_1.nonEmptyArray.of(noFieldMsg));
        }
        var notObjectMsg = "Value must be an object, found \"" + typeof valueObject + "\" instead";
        return ValidationFn.failure(NonEmptyArray_1.nonEmptyArray.of(notObjectMsg));
    };
    return FieldD;
}(Decoder));
var OptionalFieldD = /** @class */ (function (_super) {
    __extends(OptionalFieldD, _super);
    function OptionalFieldD(field, decoder) {
        var _this = _super.call(this) || this;
        _this.decoder = decoder;
        _this.field = field;
        return _this;
    }
    OptionalFieldD.prototype.run = function (valueObject) {
        var _this = this;
        if (isObject(valueObject)) {
            if (this.field in valueObject) {
                return Decoder.map(OptionFn.some, this.decoder)
                    .run(valueObject[this.field])
                    .mapFailure(function (errors) { return errors.map(function (error) { return error + " at key \"" + _this.field + "\""; }); });
            }
            return ValidationFn.success(OptionFn.none);
        }
        var notObjectMsg = "Value must be an object, found \"" + typeof valueObject + "\" instead";
        return ValidationFn.failure(NonEmptyArray_1.nonEmptyArray.of(notObjectMsg));
    };
    return OptionalFieldD;
}(Decoder));
var IndexD = /** @class */ (function (_super) {
    __extends(IndexD, _super);
    function IndexD(indexN, decoder) {
        var _this = _super.call(this) || this;
        _this.decoder = decoder;
        _this.indexN = indexN;
        return _this;
    }
    IndexD.prototype.run = function (valueArray) {
        var _this = this;
        if (isArray(valueArray)) {
            if (valueArray[this.indexN]) {
                return this.decoder
                    .run(valueArray[this.indexN])
                    .mapFailure(function (errors) { return errors.map(function (error) { return error + " at index #" + _this.indexN; }); });
            }
            var noIndexMsg = "Array missing value at index #" + this.indexN;
            return ValidationFn.failure(NonEmptyArray_1.nonEmptyArray.of(noIndexMsg));
        }
        var notArrayMsg = "Value must be an array, found \"" + typeof valueArray + "\" instead";
        return ValidationFn.failure(NonEmptyArray_1.nonEmptyArray.of(notArrayMsg));
    };
    return IndexD;
}(Decoder));
var MapD = /** @class */ (function (_super) {
    __extends(MapD, _super);
    function MapD(fn, decoder) {
        var _this = _super.call(this) || this;
        _this.decoder = decoder;
        _this.fn = fn;
        return _this;
    }
    MapD.prototype.run = function (value) {
        return this.decoder.run(value).map(this.fn);
    };
    return MapD;
}(Decoder));
var Map2D = /** @class */ (function (_super) {
    __extends(Map2D, _super);
    function Map2D(fn, decoderA, decoderB) {
        var _this = _super.call(this) || this;
        _this.decoderA = decoderA;
        _this.decoderB = decoderB;
        _this.fn = fn;
        return _this;
    }
    Map2D.prototype.run = function (value) {
        var decodedA = this.decoderA.run(value);
        var decodedB = this.decoderB.run(value);
        var fn = this.fn;
        return decodedA.fold(function (errorsA) { return decodedB.fold(function (errorsB) { return ValidationFn.failure(errorsA.concat(errorsB)); }, function () { return ValidationFn.failure(errorsA); }); }, function (successA) { return decodedB.fold(function (errorsB) { return ValidationFn.failure(errorsB); }, function (successB) { return ValidationFn.success(fn(successA, successB)); }); });
    };
    return Map2D;
}(Decoder));
var Map3D = /** @class */ (function (_super) {
    __extends(Map3D, _super);
    function Map3D(fn, decoderA, decoderB, decoderC) {
        var _this = _super.call(this) || this;
        _this.decoderA = decoderA;
        _this.decoderB = decoderB;
        _this.decoderC = decoderC;
        _this.fn = fn;
        return _this;
    }
    Map3D.prototype.run = function (value) {
        var decodedA = this.decoderA.run(value);
        var decodedB = this.decoderB.run(value);
        var decodedC = this.decoderC.run(value);
        var fn = this.fn;
        return decodedA.fold(function (errorsA) { return decodedB.fold(function (errorsB) { return decodedC.fold(function (errorsC) { return ValidationFn.failure(errorsA.concat(errorsB).concat(errorsC)); }, function () { return ValidationFn.failure(errorsA.concat(errorsB)); }); }, function () { return decodedC.fold(function (errorsC) { return ValidationFn.failure(errorsA.concat(errorsC)); }, function () { return ValidationFn.failure(errorsA); }); }); }, function (successA) { return decodedB.fold(function (errorsB) { return decodedC.fold(function (errorsC) { return ValidationFn.failure(errorsB.concat(errorsC)); }, function () { return ValidationFn.failure(errorsB); }); }, function (successB) { return decodedC.fold(function (errorsC) { return ValidationFn.failure(errorsC); }, function (successC) { return ValidationFn.success(fn(successA, successB, successC)); }); }); });
    };
    return Map3D;
}(Decoder));
var NullD = /** @class */ (function (_super) {
    __extends(NullD, _super);
    function NullD(value) {
        var _this = _super.call(this) || this;
        _this.value = value;
        return _this;
    }
    NullD.prototype.run = function (value) {
        if (value === null) {
            return ValidationFn.success(this.value);
        }
        var message = "Value must be a null, found \"" + typeof value + "\" instead";
        return ValidationFn.failure(NonEmptyArray_1.nonEmptyArray.of(message));
    };
    return NullD;
}(Decoder));
var NumberD = /** @class */ (function (_super) {
    __extends(NumberD, _super);
    function NumberD() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    NumberD.prototype.run = function (value) {
        if (typeof value === 'number') {
            return ValidationFn.success(value);
        }
        var message = "Value must be a number, found \"" + typeof value + "\" instead";
        return ValidationFn.failure(NonEmptyArray_1.nonEmptyArray.of(message));
    };
    return NumberD;
}(Decoder));
var OneOfD = /** @class */ (function (_super) {
    __extends(OneOfD, _super);
    function OneOfD(decoders) {
        var _this = _super.call(this) || this;
        _this.decoders = decoders;
        return _this;
    }
    OneOfD.prototype.run = function (value) {
        var firstResult = this.decoders.head.run(value);
        if (firstResult.isSuccess()) {
            return firstResult;
        }
        return this.decoders.tail.reduce(function (decoded, decoder) {
            if (decoded.isSuccess()) {
                return decoded;
            }
            return decoder.run(value);
        }, firstResult);
    };
    return OneOfD;
}(Decoder));
var StringD = /** @class */ (function (_super) {
    __extends(StringD, _super);
    function StringD() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    StringD.prototype.run = function (value) {
        if (typeof value === 'string') {
            return ValidationFn.success(value);
        }
        var message = "Value must be a string, found \"" + typeof value + "\" instead";
        return ValidationFn.failure(NonEmptyArray_1.nonEmptyArray.of(message));
    };
    return StringD;
}(Decoder));
var SucceedD = /** @class */ (function (_super) {
    __extends(SucceedD, _super);
    function SucceedD(value) {
        var _this = _super.call(this) || this;
        _this.value = value;
        return _this;
    }
    SucceedD.prototype.run = function () {
        return ValidationFn.success(this.value);
    };
    return SucceedD;
}(Decoder));
var UnknownD = /** @class */ (function (_super) {
    __extends(UnknownD, _super);
    function UnknownD() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    UnknownD.prototype.run = function (value) {
        return ValidationFn.success(value);
    };
    return UnknownD;
}(Decoder));
function isArray(value) {
    return Array.isArray(value);
}
function isObject(value) {
    return typeof value === 'object';
}
function pipe(value, fn) {
    return fn(value);
}
